@@ -67,23 +67,31 @@ TEST_NON_USER_AUDIENCES = [
 
 VALID_TOKEN_URLS = [
     "https://sts.googleapis.com",
+    "https://sts.mtls.googleapis.com",
     "https://us-east-1.sts.googleapis.com",
+    "https://us-east-1.sts.mtls.googleapis.com",
     "https://US-EAST-1.sts.googleapis.com",
     "https://sts.us-east-1.googleapis.com",
     "https://sts.US-WEST-1.googleapis.com",
     "https://us-east-1-sts.googleapis.com",
     "https://US-WEST-1-sts.googleapis.com",
+    "https://US-WEST-1-sts.mtls.googleapis.com",
     "https://us-west-1-sts.googleapis.com/path?query",
     "https://sts-us-east-1.p.googleapis.com",
+    "https://sts-us-east-1.p.mtls.googleapis.com",
 ]
 INVALID_TOKEN_URLS = [
     "https://iamcredentials.googleapis.com",
+    "https://mtls.iamcredentials.googleapis.com",
     "sts.googleapis.com",
+    "mtls.sts.googleapis.com",
+    "mtls.googleapis.com",
     "https://",
     "http://sts.googleapis.com",
     "https://st.s.googleapis.com",
     "https://us-eas\t-1.sts.googleapis.com",
     "https:/us-east-1.sts.googleapis.com",
+    "https:/us-east-1.mtls.sts.googleapis.com",
     "https://US-WE/ST-1-sts.googleapis.com",
     "https://sts-us-east-1.googleapis.com",
     "https://sts-US-WEST-1.googleapis.com",
@@ -95,16 +103,20 @@ INVALID_TOKEN_URLS = [
     "hhttps://us-east-1.sts.googleapis.com",
     "https://us- -1.sts.googleapis.com",
     "https://-sts.googleapis.com",
+    "https://-mtls.googleapis.com",
     "https://us-east-1.sts.googleapis.com.evil.com",
     "https://sts.pgoogleapis.com",
     "https://p.googleapis.com",
     "https://sts.p.com",
+    "https://sts.p.mtls.com",
     "http://sts.p.googleapis.com",
     "https://xyz-sts.p.googleapis.com",
     "https://sts-xyz.123.p.googleapis.com",
     "https://sts-xyz.p1.googleapis.com",
     "https://sts-xyz.p.foo.com",
     "https://sts-xyz.p.foo.googleapis.com",
+    "https://sts-xyz.mtls.p.foo.googleapis.com",
+    "https://sts-xyz.p.mtls.foo.googleapis.com",
 ]
 VALID_SERVICE_ACCOUNT_IMPERSONATION_URLS = [
     "https://iamcredentials.googleapis.com",
